import k3d
import numpy as np
import pygmsh

from ..experiments.calculate_absorption import calculate_absorption
from ..experiments.relax import minimize_gibbs_free_energy
from ..helpers.math import h_hom, h_CPW, h_strip

B_ext_color = 0xED4B42


class ExperimentalSetup:

    def __init__(self, sample, name="my_experiment"):
        self.sample = sample
        self._Bext = np.zeros(3 * sample.nx)
        self.name = name
        self._antenna = None
        self._modes_available = False

    @property
    def Bext(self):
        return self._Bext

    @Bext.setter
    def Bext(self, value):
        if self.sample.mesh == None:
            print("This sample does not have a mesh yet. You cannot set any external field for it.")
        else:
            if np.array_equal(self.sample.xyz.shape, np.shape(value)):
                self._Bext = value
            elif np.shape(value) == (3,):
                self._Bext = np.array([value for i in range(self.sample.nx)])
            else:
                print("You are trying to set an external field which does not match the shape of the mesh.")

    @property
    def antenna(self):
        return self.antenna

    @antenna.setter
    def antenna(self, value):
        if isinstance(value, MicrowaveAntenna):
            self._antenna = value
        else:
            raise ValueError(f'Input {value} is not of type "MicrowaveAntenna"')

    def show(self, comp="vec", scale=5, show_antenna=True):
        plot = k3d.plot()
        plt_mesh = k3d.mesh(np.float32(self.sample.mesh.points), np.uint32(self.sample.mesh.get_cells_type("triangle")))
        plt_mesh.wireframe = True
        plt_mesh.color = 0xaaaaaa
        plt_mesh.opacity = 1
        plot += plt_mesh

        if np.any(self._Bext) != 0:

            hx = self._Bext.T[0]
            hy = self._Bext.T[1]
            hz = self._Bext.T[2]
            h_abs = np.sqrt(hx ** 2 + hy ** 2 + hz ** 2)
            colors = np.repeat(np.array([(np.uint32(B_ext_color), np.uint32(B_ext_color))]), self.sample.nx)
            if comp == "vec":
                m_mesh = k3d.vectors(np.float32(self.sample.mesh.points),
                                     np.float32(scale * (np.vstack([hx, hy, hz])).T),
                                     head_size=2.5 * scale * np.mean(h_abs),
                                     line_width=0.05 * scale * np.mean(h_abs), colors=colors)
            else:
                comp_dict = {"x": hx, "y": hy, "z": hz, "phi": np.arctan2(hy, hx),
                             "abs": h_abs}
                color_dict = {"abs": k3d.colormaps.matplotlib_color_maps.Inferno,
                              "x": k3d.colormaps.matplotlib_color_maps.RdBu,
                              "y": k3d.colormaps.matplotlib_color_maps.RdBu,
                              "z": k3d.colormaps.matplotlib_color_maps.RdBu,
                              "phi": k3d.colormaps.matplotlib_color_maps.Twilight_shifted}

                # TODO check if comp is valid key
                m_mesh = k3d.mesh(np.float32(self.sample.mesh.points),
                                  np.uint32(self.sample.mesh.get_cells_type("triangle")), attribute=comp_dict[comp],
                                  color_map=color_dict[comp], interpolation=False, side="double", flat_shading=True,
                                  name="$m_{{}}$".format(comp))
            plot += m_mesh

        else:
            anchor = 1.5 * np.max(self.sample.xyz)
            plt_text = k3d.text2d('B_{ext} = 0',
                                  position=[0, 0],
                                  color=B_ext_color, size=1)
            plot += plt_text

        if self._antenna is not None and show_antenna:
            span = np.max(np.abs(self.sample.xyz.T[0])) * 2
            antenna_mesh = self._antenna.get_mesh(span)
            plot += k3d.mesh(np.float32(antenna_mesh.points), np.uint32(antenna_mesh.get_cells_type("triangle")),
                             color=0xFFD700, side="double")

        plot.display()

    def relax(self, tol=1e-12, continue_least_with_squares=False, return_last=False):
        # TODO check if sample has magnetization
        if self.sample.mag is None:
            print("Your sample does not have a magnetization field yet.")
        else:
            self.sample.mag = minimize_gibbs_free_energy(self.sample, self._Bext, tol_cg=tol, return_last=return_last,
                                                         continue_least_with_squares=continue_least_with_squares)

    def eigenmodes(self, kmin=-40e6, kmax=40e6, Nk=81, num_modes=20, k=None, no_dip=False, num_cpus=1, save_modes=True,
                   save_local=False, save_magphase=False):
        if self.sample.mag is None:
            print("Your sample does not have a magnetization field yet.")
        else:
            dispersion = self.sample.eigensolver(self.sample, self.Bext, self.name, kmin, kmax, Nk, num_modes, k,
                                                 no_dip, num_cpus, save_modes, save_local, save_magphase)
            self._modes_available = save_modes
            self.dispersion = dispersion
            return dispersion

    def absorption(self, auto_eigenmodes=False, fmin=0, fmax=30, Nf=200, k=None):
        if self.sample.mag is None:
            print("Your sample does not have a magnetization field yet.")
        else:
            if self._modes_available is False and auto_eigenmodes is False:
                print("You did not calculate any eigenmodes to calculate an absorption from. "
                      "Please run .eigenmodes(..., save_modes=True) in this ExperimentalSetup "
                      "first or set auto_eigenmodes=True when calling absorption().")

            if self._modes_available is False and auto_eigenmodes is True:
                print("Calculating eigenmodes first.")
                self.dispersion = self.eigenmodes(save_modes=True, k=k)
                print("Calculating absorption.")
                power_map = calculate_absorption(self.sample, self.name, self.dispersion, self._antenna, fmin, fmax, Nf,
                                                 k)
                print("Done.")
                return power_map
            if self._modes_available is True and auto_eigenmodes is True:
                print("Modes are already calculated. Use auto_eigenmodes=False instead.")

            if self._modes_available is True and auto_eigenmodes is False:
                print("Calculating absorption.")
                power_map = calculate_absorption(self.sample, self.name, self.dispersion, self._antenna, fmin, fmax, Nf,
                                                 k)
                print("Done.")
                return power_map


def create_experimental_setup(sample, name: str = "my_experiment") -> ExperimentalSetup:
    """Creates an experimental setup around a given sample.

    Parameters
    ----------
    sample : AbstractSample
        Sample object to create ExperimentalSetup around.

    name : str
        Name of the experimental setup.

    Returns
    -------
    ExperimentalSetup
    """

    return ExperimentalSetup(sample, name)


class MicrowaveAntenna:

    def __init__(self):
        self.type = None


class StriplineAntenna(MicrowaveAntenna):

    def __init__(self, width, spacing):
        self.type = "stripline"
        self.width = width
        self.spacing = spacing
        self.k_spectrum = h_strip

    def get_mesh(self, span):
        x0 = [-span / 2, self.spacing - 2.5, -self.width / 2]
        with pygmsh.occ.Geometry() as geom:
            geom.add_box(x0, [span, 5, self.width], self.width)
            mesh = geom.generate_mesh()
        return mesh

    def get_microwave_field_function(self, xyz):
        x_ = xyz.T[0]
        antenna_func = lambda k: h_strip(k, 1, self.width, np.abs(x_ - self.spacing))
        antenna_field_func = lambda k: np.array([np.zeros_like(x_), antenna_func(k), antenna_func(k)]).flatten()
        return antenna_field_func


class CPWAntenna(MicrowaveAntenna):

    def __init__(self, width, gap, spacing):
        self.type = "stripline"
        self.width = width
        self.gap = gap
        self.spacing = spacing
        self.k_spectrum = h_CPW

    def get_mesh(self, span):
        with pygmsh.occ.Geometry() as geom:
            geom.add_box([-span / 2, self.spacing - 2.5, -self.width * 1.5 - self.gap],
                         [span, 5, self.width],
                         self.width)
            geom.add_box([-span / 2, self.spacing - 2.5, -self.width / 2],
                         [span, 5, self.width],
                         self.width)
            geom.add_box([-span / 2, self.spacing - 2.5, self.width / 2 + self.gap],
                         [span, 5, self.width],
                         self.width)
            mesh = geom.generate_mesh()
        return mesh

    def get_microwave_field_function(self, xyz):
        x_ = xyz.T[0]
        antenna_func = lambda k: h_CPW(k, 1, self.width, self.gap, np.abs(x_ - self.spacing))
        antenna_field_func = lambda k: np.array([np.zeros_like(x_), antenna_func(k), antenna_func(k)]).flatten()
        return antenna_field_func


class HomogeneousAntenna(MicrowaveAntenna):

    def __init__(self, theta=0, phi=0):
        self.type = "homogeneous"
        self.theta = theta / 180 * np.pi
        self.phi = phi / 180 * np.pi

    def get_mesh(self):
        with pygmsh.occ.Geometry() as geom:
            # geom.add_box(x0, [span, 5, self.width], self.width)
            mesh = geom.generate_mesh()
        return mesh

    def get_microwave_field_function(self, xyz):
        x_ = xyz.T[0]
        antenna_func = lambda k: h_hom(k, np.ones_like(x_))
        antenna_field_func = lambda k: np.array([np.sin(self.theta) * np.cos(self.phi) * antenna_func(k),
                                                 np.sin(self.theta) * np.sin(self.phi) * antenna_func(k),
                                                 np.cos(self.theta) * antenna_func(k)]).flatten()
        return antenna_field_func


class CurrentLoopAntenna(MicrowaveAntenna):

    def __init__(self, width, radius):
        self.type = "currentloop"
        self.width = width
        self.radius = radius

    def get_mesh(self, span):
        with pygmsh.occ.Geometry() as geom:
            geom.characteristic_length_min = self.radius * np.pi / 30
            geom.characteristic_length_max = self.radius * np.pi / 20
            cylinder1 = geom.add_cylinder([0, 0, -self.width / 2], [0, 0, self.width], self.radius + 2.5, 2 * np.pi)
            cylinder2 = geom.add_cylinder([0, 0, -self.width / 2], [0, 0, self.width], self.radius - 2.5, 2 * np.pi)
            geom.boolean_difference(cylinder1, cylinder2)
            mesh = geom.generate_mesh()
        return mesh

    def get_microwave_field_function(self, xyz):
        x_ = xyz.T[0]
        y_ = xyz.T[1]
        rho_ = np.sqrt(x_ ** 2 + y_ ** 2)
        phi_ = np.arctan2(y_, x_)
        antenna_func = lambda k: h_strip(k, 1, self.width, np.abs(rho_ - self.radius))

        # antenna_field_func = lambda k: np.array([np.zeros_like(x_), np.zeros_like(x_), antenna_func(k)]).flatten()

        antenna_field_func = lambda k: np.array(
            [antenna_func(k) * np.cos(phi_), antenna_func(k) * np.sin(phi_), antenna_func(k)]).flatten()
        return antenna_field_func
