from abc import ABC, abstractmethod

import k3d
import meshio
import numpy as np

from .fem_core.cythoncore import get_matrices
from .operators import ExchangeOperator, DipolarOperator, UniAxialAnisotropyOperator, BulkDMIOperator, \
    InterfacialDMIOperator
from ..experiments.eigen import calculate_normal_modes, not_implemented_eigensolver
from ..helpers.io import check_mesh_shape
from ..helpers.io import get_mesh_dimension
from ..helpers.math import normalize_single_3d_vec


class AbstractSample(ABC):

    def __init__(self, name, geometry, magnetic_order, dim):
        self.name = name
        self._dim = dim
        self._geometry = geometry
        self._magnetic_order = magnetic_order

        # geometric parameters
        self.geom = None
        self.mesh = None
        self.nx = None
        self.nb = None
        self.scale = 1e-9
        self.xyz = None

        self.get_eigensolver()
        self.initialize_material_params()

    @abstractmethod
    def get_eigensolver(self):
        """This method should be implemented by the FerromagneticMixin or
        AntiferromagneticMixin class of the concrete sample."""

    @abstractmethod
    def initialize_material_params(self):
        """This method should be implemented by the FerromagneticMixin or
        AntiferromagneticMixin class of the concrete sample."""

    @abstractmethod
    def get_magnetic_tensors(self):
        """This method should be implemented by the FerromagneticMixin or
        AntiferromagneticMixin class of the concrete sample."""

    @abstractmethod
    def show(self):
        """This method should be implemented by the FerromagneticMixin or
        AntiferromagneticMixin class of the concrete sample."""

    def set_geom(self, mesh):

        input_mesh_dim = get_mesh_dimension(mesh)
        if input_mesh_dim != self._dim:
            raise ValueError(f"The dimension of the input mesh ({input_mesh_dim}) does "
                             f"not fit your chosen geometry ({self._geometry}).")

        if not check_mesh_shape(mesh, self._dim):
            raise ValueError(
                f"The input mesh is not properly shaped occording two your chosen geometry "
                f"({self._geometry}). For waveguide samples, all nodes should be in the xy "
                f"plane. For layer samples, all nodes should be on the y axis.")

        print("Setting geometry and calculating discretized differential operators on mesh.")

        self.mesh = mesh
        self.xyz = self.mesh.points
        self.nx = len(self.xyz)

        self.boundary_nodes, \
        self.nv, \
        self.dA, \
        self.pang, \
        self.belm, \
        self.xc, \
        self.poisson, \
        self.laplace, \
        self.div_x, \
        self.div_y, \
        self.grad_x, \
        self.grad_y = get_matrices(self.xyz,
                                   self.mesh.get_cells_type("triangle"),
                                   self.Aex,
                                   self.Msat,
                                   self.scale)
        self.nb = len(self.boundary_nodes)

        # correcting wrong sign from partial integration
        self.xc *= -1
        self.poisson *= -1
        self.laplace *= -1
        self.div_x *= -1
        self.div_y *= -1
        self.grad_x *= -1
        self.grad_y *= -1
        self.get_magnetic_tensors()
        print("Done.")

    def read_mesh(self, fname):
        try:
            mesh = meshio.read(fname)
        except ReadError:
            print("Could not read this mesh file.")
        self.set_geom(mesh)


class FerromagnetMixin:
    def initialize_material_params(self):

        # material parameters
        self.Msat = 796e3
        self.Aex = 13e-12
        self.gamma = 1.760859644e11
        self.alpha = 0.01
        self.Didmi = 0.0
        self.Dbulk = 0.0

        self.Ku1 = 0.0
        self.Ku2 = 0.0
        self.k_phi = 0.0
        self.k_theta = 0.0
        self.e_u = np.array([0, 0, 1])  # uniaxial anistropy direction(s)

        self.Kc1 = 0.0
        self.v1Kc = np.array([1, 0, 0])
        self.v2Kc = np.array([0, 1, 0])
        self.v3Kc = np.array([0, 0, 1])

        # mag is hidden
        self._mag = None

    @property
    def mag(self):
        return self._mag

    @mag.setter
    def mag(self, value):
        if self.mesh == None:
            print("This sample does not have a mesh yet. You cannot set any magnetization for it.")
        else:
            if self.xyz.shape == np.shape(value):
                # TODO make unitary
                self._mag = normalize_single_3d_vec(value)
            elif np.shape(value) == (3,):
                self._mag = normalize_single_3d_vec(np.array([value for i in range(self.nx)]))
            else:
                print("You are trying to set a magnetization field which does not match the shape of the mesh. \n" + \
                      "mesh shape: {}\n".format(self.xyz.shape) + \
                      "value shape: {}\n".format(np.shape(value)))

    def get_magnetic_tensors(self):
        self.N_exc = ExchangeOperator(self)
        self.N_dip = DipolarOperator(self)
        self.N_uni = UniAxialAnisotropyOperator(self)
        self.N_DMI = BulkDMIOperator(self)
        self.N_iDMI = InterfacialDMIOperator(self)

    def show(self, show_node_labels=False, comp="vec", scale=5, show_mag=True):
        if self.mesh == None:
            print("This sample does not have a mesh yet.")
        else:
            plot = k3d.plot(lighting=0)
            plt_mesh = k3d.mesh(np.float32(self.mesh.points), np.uint32(self.mesh.get_cells_type("triangle")))
            plt_mesh.wireframe = True
            plt_mesh.color = 0xaaaaaa
            plt_mesh.opacity = 1
            plot += plt_mesh

            if show_node_labels:
                text_obj = k3d.vectors(np.float32(self.mesh.points), 0 * np.float32(self.mesh.points),
                                       labels=[str(i) for i in range(len(self.mesh.points))])
                plot += text_obj

            if np.all(self._mag) is not None and show_mag:
                mx = self._mag.T[0]
                my = self._mag.T[1]
                mz = self._mag.T[2]

                if comp == "vec":
                    m_mesh = k3d.vectors(np.float32(self.mesh.points), np.float32(scale * (np.vstack([mx, my, mz])).T),
                                         head_size=2.5 * scale,
                                         line_width=0.05 * scale)
                else:
                    comp_dict = {"x": mx, "y": my, "z": mz, "phi": np.arctan2(my, mx),
                                 "abs": np.sqrt(mx ** 2 + my ** 2 + mz ** 2)}
                    color_dict = {"abs": k3d.colormaps.matplotlib_color_maps.Inferno,
                                  "x": k3d.colormaps.matplotlib_color_maps.RdBu,
                                  "y": k3d.colormaps.matplotlib_color_maps.RdBu,
                                  "z": k3d.colormaps.matplotlib_color_maps.RdBu,
                                  "phi": k3d.colormaps.matplotlib_color_maps.Twilight_shifted}

                    # TODO check if comp is valid key
                    m_mesh = k3d.mesh(self.mesh.points, self.mesh.get_cells_type("triangle"), attribute=comp_dict[comp],
                                      color_map=color_dict[comp], interpolation=False, side="double", flat_shading=True,
                                      name="$m_{{}}$".format(comp))
                plot += m_mesh
            plot.display()

    def _repr_html_(self):
        return f"""<p><h3>Sample: {self.name}</h3></p>
               <p><table align="left">
              <tr>
                <th>Material parameter</th>
                <th>Attribute name</th>
                <th>Value</th>
              </tr>
              <tr>
                <td>Saturation magnetization </td><td><tt>Msat</tt></td>
                <td>{self.Msat} A/m</td>
              </tr>
              <tr>
                <td>Exchange stiffness constant  </td><td><tt>Aex</tt></td>
                <td>{self.Aex} J/m</td>
              </tr>
              <tr>
                <td>Gyromagnetic ratio  </td><td><tt>gamma</tt></td>
                <td>{self.gamma} rad Hz/T</td>
              </tr>
              <tr>
                <td>Gilbert damping parameter  </td><td><tt>alpha</tt></td>
                <td>{self.alpha}</td>
              </tr>
              <tr>
                <td>Unaxial anistropy constant  </td><td><tt>Ku1</tt></td>
                <td>{self.Ku1} J/m<sup>3</sup></td>
              </tr>
              <tr>
                <td>Unaxial anistropy angles  </td><td><tt>k_phi</tt><br><tt>k_theta</tt></td>
                <td>{self.k_phi}°<br>{self.k_theta}°</td>
              </tr>
              <tr>
                <td>Cubic anistropy constant</td><td><tt>Kc1</tt></td>
                <td>{self.Kc1} J/m<sup>3</sup></td>
              </tr>
                <tr>
                    <td>Cubic anistropy axes</td><td><tt>v1Kc</tt><br><tt>v2Kc</tt></td>
                    <td>({self.v1Kc[0]}, {self.v1Kc[1]}, {self.v1Kc[2]})<br>({self.v2Kc[0]}, {self.v2Kc[1]}, {self.v2Kc[2]})</td>
                  </tr>
              <tr>
                <td>Bulk DMI constant  </td><td><tt>Dbulk</tt></td>
                <td>{self.Dbulk} J/m<sup>2</sup></td>
              </tr>
             <tr>
                <td>Interfacial DMI constant  </td><td><tt>Didmi</tt></td>
                <td>{self.Didmi} J/m<sup>2</sup></td>
              </tr>
              <tr>
                 <td>&nbsp;   </td><td></td>
                <td>&nbsp;  </td>
              </tr>
              <tr>
                <th>Mesh parameter  </th><th>Attribute name</th>
                <th>Value</th>
              </tr>
              <tr>
                <td>Mesh scaling</td><td><tt>scale</tt></td>
                <td>{self.scale}</td>
              </tr>
               <tr>
                <td>Number of nodes</td><td><tt>nx</tt></td>
                <td>{self.nx}</td>
              </tr>
               <tr>
                <td>Number of boundary nodes</td><td><tt>nb</tt></td>
                <td>{self.nb}</td>
              </tr>

            </table>
            </p>"""


class AntiferromagnetMixin:
    def initialize_material_params(self):
        self.params = ["a", "couple", "of", "antiferromagnetic", "parameters"]

        self._mag1 = None
        self._mag2 = None

    @property
    def mag1(self):
        return self._mag1

    @mag1.setter
    def mag1(self, value):
        self._mag1 = value

    @property
    def mag2(self):
        return self._mag2

    @mag2.setter
    def mag2(self, value):
        self._mag2 = value

    def get_magnetic_tensors(self):
        print("Antiferromagnetic tensors are not implemented yet.")

    def show(self):
        pass


class PropagatingWaveEigensolverMixin:
    def get_eigensolver(self):
        self.eigensolver = calculate_normal_modes


class ConfinedWaveEigensolverMixin:
    def get_eigensolver(self):
        self.eigensolver = not_implemented_eigensolver


class ConfinedSampleFM(ConfinedWaveEigensolverMixin, FerromagnetMixin, AbstractSample):
    def description(self):
        print("I am a ferromagnetic confined sample.")


class ConfinedSampleAFM(ConfinedWaveEigensolverMixin, AntiferromagnetMixin, AbstractSample):
    def description(self):
        print("I am a antiferromagnetic confined sample.")


class WaveguideSampleFM(PropagatingWaveEigensolverMixin, FerromagnetMixin, AbstractSample):
    def description(self):
        print("I am a ferromagnetic waveguide sample.")


class WaveguideSampleAFM(PropagatingWaveEigensolverMixin, AntiferromagnetMixin, AbstractSample):
    def description(self):
        print("I am a antiferromagnetic waveguide sample.")


class LayerSampleFM(PropagatingWaveEigensolverMixin, FerromagnetMixin, AbstractSample):
    def description(self):
        print("I am a ferromagnetic layer sample.")


class LayerSampleAFM(PropagatingWaveEigensolverMixin, AntiferromagnetMixin, AbstractSample):
    def description(self):
        print("I am a antiferromagnetic layer sample.")


def create_sample(name="my_sample", geometry="waveguide", magnetic_order="FM"):
    """Create a sample with certain geometry and magnetic order.

    Parameters
    ----------
    geometry : str, {"confined", "waveguide", "layer"}
        Geometry of the sample (default is `"waveguide"`).
    magnetic_order : str {"FM", "AFM"}, optional
        Magnetic order of the material. Can be either ferromagnetic (`"FM"`)
        or antiferromagnetic (`"AFM"`). The default is `"FM"`.


    Returns
    -------
    sample : AbstractSample

    """

    ALLOWED_GEOMETRIES = {"confined", "waveguide", "layer"}
    ALLOWED_MAGNETIC_ORDERS = {"FM", "AFM"}

    # Which sample type to choose depending on magnetic_order and geometry.

    SAMPLE_DICT = {("confined", "FM"): ConfinedSampleFM(name, geometry, magnetic_order, 3),
                   ("confined", "AFM"): ConfinedSampleAFM(name, geometry, magnetic_order, 3),
                   ("waveguide", "FM"): WaveguideSampleFM(name, geometry, magnetic_order, 2),
                   ("waveguide", "AFM"): WaveguideSampleAFM(name, geometry, magnetic_order, 2),
                   ("layer", "FM"): LayerSampleFM(name, geometry, magnetic_order, 1),
                   ("layer", "AFM"): LayerSampleAFM(name, geometry, magnetic_order, 1),
                   }

    # check if supplied parameters are valid.

    if geometry not in ALLOWED_GEOMETRIES:
        raise ValueError(f'Invalid geometry value "{geometry}"; should be "confined", '
                         '"waveguide" or "layer".')

    if magnetic_order not in ALLOWED_MAGNETIC_ORDERS:
        raise ValueError(f'Invalid magnetic_order value "{magnetic_order}"; should be "FM", or "AFM".')

    return SAMPLE_DICT[(geometry, magnetic_order)]
