import multiprocessing as mp
import os
from functools import partial

import meshio
import numpy as np
import pandas as pd
import tqdm
from scipy.constants import mu_0
from scipy.sparse import csc_matrix
from scipy.sparse.linalg import spilu, eigs

from ..core.fem_core.cythoncore import rotation_matrix_py
from ..core.operators import cross_operator, TotalDynMat, SuperLUInv, InvTotalDynMat
from ..helpers.math import flattened_mesh_vec_scalar_product, diag_matrix_from_vec


def not_implemented_eigensolver(sample, Bext, exp_name, kmin=-40e6, kmax=40e6, Nk=81, num_modes=20, k=None, no_dip=False,num_cpus=1, save_modes=False,save_local=False, save_magphase=False):
    print("This eigensolver will be implemented in a future release.")


def test_rotation(mag, rotation, nx):
    """
    Test of rotation matrix and equilibrium vector match, i.e. rotation rotates mag into the z direction.
    """
    test_ez = rotation.dot(mag)
    e_z = np.concatenate([np.zeros(2*nx), np.ones(nx)])
    if not np.allclose(test_ez, e_z):
        raise ValueError("Rotation matrix and equilibrium files don't belong together. "
                         "Please create a new rotation matrix.")


def calculate_normal_modes(sample, Bext, exp_name, kmin=-40e6, kmax=40e6, Nk=81, num_modes=20, k=None, no_dip=False,num_cpus=1, save_modes=False,save_local=False, save_magphase=False):
    """
        2D version
    """

    modes_path = "./{}/{}/mode-profiles".format(sample.name,exp_name)
    if not os.path.exists(modes_path):
        os.makedirs(modes_path)

    # exchange operator
    N_exc = sample.N_exc
    N_dip = sample.N_dip
    N_uni = sample.N_uni
    N_DMI = sample.N_DMI
    N_iDMI = sample.N_iDMI


    N_iDMI.set_k(0)
    N_DMI.set_k(0)
    N_exc.set_k(0)
    N_dip.set_k(0)

    mu0_cross = cross_operator(sample.nx)

    mag = sample.mag.T.flatten()
    rotation = rotation_matrix_py(mag)
    test_rotation(mag, rotation, sample.nx)


    nx = sample.nx
    hext = Bext.T.flatten() / (mu_0 * sample.Msat)
    hexc = -N_exc.dot(mag)
    hdip = -N_dip.dot(mag)
    hDMI = -N_DMI.dot(mag)
    hiDMI = -N_iDMI.dot(mag)
    hani = -N_uni.dot(mag)
    #hcub = N_cub.nonlinear_field(mag)

    if no_dip:
        heff = hexc + hext  + hani  + hDMI + hdip  + hiDMI # ++ hcub

    else:
        heff = hexc + hdip + hext + hani  + hDMI + hiDMI #+ hcub

    h0 = flattened_mesh_vec_scalar_product(heff, mag)

    rotation = rotation[:2 * nx, :]
    rotation_T = rotation.transpose()


    # create k and convert to mesh units
    if k == None:
        k_ = np.linspace(kmin,kmax,Nk) * sample.scale
    else:
        k_ = [k * sample.scale]

    v0 = np.full(2*nx, 1+1j, dtype=complex)


    D_tot = TotalDynMat(N_exc.sparse_mat, N_dip, 1j * mu0_cross[:2 * nx, :2 * nx].dot(rotation), rotation_T)

    Lambda = mu0_cross[:2 * nx, :2 * nx]

    bndint_order = 6

    modes_per_k_partial = partial(modes_per_k, N_exc, N_uni, N_DMI, N_iDMI, h0, Lambda, rotation, rotation_T, bndint_order, D_tot, v0, no_dip, sample.scale, sample.Msat, sample.gamma, sample, num_modes, modes_path, save_modes,save_local, save_magphase)

    if num_cpus == 0 or num_cpus < -1:
        num_cpus = 1
    if num_cpus == -1:
        num_cpus = mp.cpu_count()
    if num_cpus > mp.cpu_count():
        num_cpus = mp.cpu_count()

    if num_cpus == 1:
        if mp.cpu_count() > 1:
            pass
    else:
        pass


    with mp.Pool(processes=num_cpus) as p:
        res_list = []
        with tqdm.tqdm(total=len(k_)) as pbar:
            for i, res in enumerate(p.imap_unordered(modes_per_k_partial, k_)):
                res_list.append(res)
                pbar.update()
    df_freqs = pd.concat(res_list,sort=False).sort_values("k (rad/m)").reset_index(drop=True).fillna("nan")

    return df_freqs


def modes_per_k(N_exc, N_uni, N_DMI, N_iDMI, h0, Lambda, rotation, rotation_T, bndint_order, D_tot, v0, no_dip, scale, Msat, gamma,sample, num_modes, modes_path, save_modes, save_local, save_magphase, k):
    N_exc.set_k(k)
    N_DMI.set_k(k)
    N_iDMI.set_k(k)

    N_nodip_k_mat = N_exc.sparse_mat + N_uni.sparse_mat + N_DMI.sparse_mat + N_iDMI.sparse_mat
    N_nodip_k_mat += diag_matrix_from_vec(np.tile(h0, 3))
    dyn_mat_k = 1j * Lambda.dot(rotation).dot(N_nodip_k_mat).dot(rotation_T)


    dyn_mat_k_csc = csc_matrix(dyn_mat_k)

    if no_dip:
        freq, eigenvectors = eigs(dyn_mat_k_csc, v0=v0, which="LM", k=num_modes * 2, tol=1e-3, sigma=0)
        ef = freq.real * mu_0 * Msat * gamma / 2 / np.pi * 1e-9

    else:
        D_tot.set_k(k, N_nodip_k_mat)
        try:
            preconditioner = SuperLUInv(spilu(csc_matrix(dyn_mat_k)))
        except:
            log.warning(
                "It was not possible to calculate the preconditioner at k = {} rad/m".format(k_m=k / conf["scale"]))
            quit()

        D_tot_inv = InvTotalDynMat(D_tot, k, preconditioner=preconditioner)
        try:
            freq, eigenvectors = eigs(D_tot, v0=v0, which="LM", k=num_modes * 2, tol=1e-3, sigma=0, OPinv=D_tot_inv)
            ef = freq.real * mu_0 * Msat * gamma / 2 / np.pi * 1e-9
        except InvertError as this_error:
            error_k = this_error.k
            error_iter = this_error.after_iter
            error_success = this_error.already_sucess

            ef = np.zeros(num_modes)
            eigenvectors = np.zeros((2*D_tot.nx, num_modes))


    ef_org = ef.copy()

    # sort out all negative frequency modes
    gtzero = ef >= 0
    eigenvectors = eigenvectors[:, gtzero]
    ef = ef[gtzero]

    # now sort modes by ascending frequency
    sortindices = np.argsort(ef)
    ef = ef[sortindices]
    eigenvectors = eigenvectors[:, sortindices]

    k_m = k / scale


    nx = sample.nx
    cells = sample.mesh.cells
    xyz = sample.xyz

    if save_modes:

        for i in range(eigenvectors.shape[1]):
            ev = eigenvectors[:, i]

            # rotate eigenvector back to lab system
            ev_lab = rotation_T.dot(ev)
            ev_lab_x = ev_lab[:nx]
            ev_lab_y = ev_lab[nx:-nx]
            ev_lab_z = ev_lab[-nx:]

            save_dict = {"Re(m_x)" : ev_lab_x.real,
                         "Im(m_x)" : ev_lab_x.imag,
                         "Re(m_y)" : ev_lab_y.real,
                         "Im(m_y)" : ev_lab_y.imag,
                         "Re(m_z)" : ev_lab_z.real,
                         "Im(m_z)": ev_lab_z.imag,
                        }

            if save_local:
                ev_1 = ev[:nx]
                ev_2 = ev[nx:]

                save_dict["Re(m_1)"] = ev_1.real
                save_dict["Im(m_1)"] = ev_1.imag
                save_dict["Re(m_2)"] = ev_2.real
                save_dict["Im(m_2)"] = ev_2.imag

                if save_magphase:
                    save_dict["Abs(m_1)"] = np.abs(ev_1)
                    save_dict["Abs(m_2)"] = np.abs(ev_2)
                    save_dict["Arg(m_1)"] = np.angle(ev_1)
                    save_dict["Arg(m_2)"] = np.angle(ev_2)

            if save_magphase:
                save_dict["Abs(m_x)"] = np.abs(ev_lab_x)
                save_dict["Abs(m_y)"] = np.abs(ev_lab_y)
                save_dict["Abs(m_z)"] = np.abs(ev_lab_z)

                save_dict["Arg(m_x)"] = np.angle(ev_lab_x)
                save_dict["Arg(m_y)"] = np.angle(ev_lab_y)
                save_dict["Arg(m_z)"] = np.angle(ev_lab_z)

            meshio.write_points_cells(filename="{}/mode_k{}radperum_{:03d}.vtk".format(modes_path,k_m * 1e-6, i),
                                      points=xyz,
                                      cells=cells,
                                      point_data=save_dict
                                      )

    else:
        pass

    df_k = pd.DataFrame(columns=["k (rad/m)"] + ["f" + str(j) + " (GHz)" for j in range(len(ef))])
    df_k.loc[0] = [k_m] + np.real(ef).tolist()
    return df_k