import time

import numpy as np
from scipy.constants import mu_0
from scipy.optimize import minimize
from scipy.sparse import csr_matrix

from ..helpers.math import flattened_mesh_vec_scalar_product, spherical_angles_to_mesh_vector


def not_implemented_minimizer(sample, Bext, tol_cg, return_last=False, continue_least_with_squares=False):
    print("This energy minimizer will be implemented in a future release.")


def energy_per_length_spherical(mag_spherical, nx, dA, h_ext, N, N_cub):
    # TODO include faster scalar product!

    # unpack
    theta = mag_spherical[:nx]
    phi = mag_spherical[-nx:]

    # convert to carthesian coordinates
    m_x = np.sin(theta) * np.cos(phi)
    m_y = np.sin(theta) * np.sin(phi)
    m_z = np.cos(theta)
    mag = np.concatenate((m_x, m_y, m_z))

    # print(mag.shape)
    # print((h_ext - 0.5 * N.dot(mag)).shape)
    # energy_per_volume = -tetramagvec_scalar_product(mag, h_ext + 0.5*N_cub.nonlinear_field(mag) - 0.5 * N.dot(mag)).real
    energy_per_volume = -flattened_mesh_vec_scalar_product(mag, h_ext - 0.5 * N.dot(mag)).real

    return (energy_per_volume * dA).sum()


def jac_spherical(mag_spherical, nx, dA, h_ext, N, N_cub):
    # TODO include faster scalar product!
    # e_per_V = -tetramagvec_scalar_product(mag, h_ext)

    # unpack
    theta = mag_spherical[:nx]
    phi = mag_spherical[-nx:]

    # convert to carthesian coordinates
    st = np.sin(theta)
    sp = np.sin(phi)
    ct = np.cos(theta)
    cp = np.cos(phi)

    m_x = np.sin(theta) * np.cos(phi)
    m_y = np.sin(theta) * np.sin(phi)
    m_z = np.cos(theta)
    mag = np.concatenate((m_x, m_y, m_z))

    h_eff = h_ext - N.dot(mag).real  # + N_cub.nonlinear_field(mag)
    # TODO include again!

    h_x = h_eff[:nx]
    h_y = h_eff[nx:-nx]
    h_z = h_eff[-nx:]

    dm00 = cp * ct
    dm01 = -sp * st
    dm10 = sp * ct
    dm11 = cp * st
    dm20 = -st
    dm21 = 0.0

    dw_theta = -1 * (h_x * dm00 + h_y * dm10 + h_z * dm20)
    dw_phi = -1 * (h_x * dm01 + h_y * dm11 + h_z * dm21)

    return np.concatenate((dw_theta, dw_phi))


def minimize_gibbs_free_energy(sample, Bext, tol_cg, return_last=False, continue_least_with_squares=False):
    # read_tmv_from_project("area.bin", conf)
    nx = sample.nx
    # area associated with each node
    dA = sample.dA
    noise = 1e-10
    h_ext = Bext.T.flatten() / (mu_0 * sample.Msat)

    m_ini = sample._mag.copy()
    m_x = m_ini.T[0]
    m_y = m_ini.T[1]
    m_z = m_ini.T[2]

    theta = np.arccos(m_z)
    phi = np.arctan2(m_y, m_x)
    m_ini_spherical = np.concatenate((theta, phi)) + noise * np.random.randn(2 * nx)

    sample.N_dip.set_k(0)
    sample.N_exc.set_k(0)
    sample.N_DMI.set_k(0)
    sample.N_iDMI.set_k(0)

    N_tot = sample.N_dip + sample.N_exc + sample.N_uni + sample.N_DMI + sample.N_iDMI
    N_cub = csr_matrix([])

    fac = (sample.Msat * sample.scale) ** 2 * mu_0
    # tol_cg = 1e-12

    print_current_spherical = lambda x: print(
        "Current energy length density: {:.15e} J/m  mx = {:.2f}  my = {:.2f}  mz = {:.2f}\r".format(
            (energy_per_length_spherical(x, nx, dA, h_ext, N_tot, N_cub) * fac),
            (np.sin(x[:nx]) * np.cos(x[-nx:])).mean(),
            (np.sin(x[:nx]) * np.sin(x[-nx:])).mean(),
            (np.cos(x[:nx])).mean()), end="")

    starting_method = "L-BFGS-B"
    print(f"Minimizing in using '{starting_method}' (tolerance {tol_cg}) ...")
    t_start = time.time()
    result = minimize(energy_per_length_spherical, m_ini_spherical, (nx, dA, h_ext, N_tot, N_cub), jac=jac_spherical,
                      callback=print_current_spherical, method=starting_method,
                      options={"ftol": tol_cg, "gtol": tol_cg})
    t_end = time.time()
    if result.success:
        theta = result.x[:nx]
        phi = result.x[-nx:]

        # convert to carthesian coordinates
        m_final = spherical_angles_to_mesh_vector(theta, phi)

        # print(m_final.shape)
        # print(sample.xyz.shape)
        print("\nSuccess!\n")
        return m_final
    else:
        print("\ndid not work :(\n")
        if not continue_least_with_squares:
            if not return_last:
                return m_ini
            else:
                theta = result.x[:nx]
                phi = result.x[-nx:]

                # convert to carthesian coordinates
                m_final = spherical_angles_to_mesh_vector(theta, phi)
                print("Returning last iteration step.")
                return m_final

        if continue_least_with_squares:
            print(f"Minimizing in using SLSQP method (tolerance {tol_cg}) ...")
            result = minimize(energy_per_length_spherical, result.x, (nx, dA, h_ext, N_tot, N_cub), jac=jac_spherical,
                              callback=print_current_spherical, method='SLSQP', options={"ftol": tol_cg})

            if result.success:
                theta = result.x[:nx]
                phi = result.x[-nx:]

                # convert to carthesian coordinates
                m_final = spherical_angles_to_mesh_vector(theta, phi)

                print("\nSuccess!\n")
                return m_final

            else:
                print("\nDid not work either.\n")

                if return_last:
                    print("\nReturning last iteration step.\n")
                    return spherical_angles_to_mesh_vector(result.x[:nx], result.x[-nx:])
                else:
                    return m_ini
