import numpy as np


def helical(xyz, Theta, chi):
    Theta = Theta / 180 * np.pi
    x = xyz.T[0]
    y = xyz.T[1]
    phi = np.arctan2(y, x)
    mx = chi * np.sin(Theta) * (-np.sin(phi))
    my = chi * np.sin(Theta) * (np.cos(phi))
    mz = np.cos(Theta) * np.ones_like(x)
    return np.array([mx, my, mz]).T


def radial(xyz, p):
    x = xyz.T[0]
    y = xyz.T[1]
    mx = p * x / np.sqrt(x ** 2 + y ** 2)
    my = p * y / np.sqrt(x ** 2 + y ** 2)
    mz = np.zeros_like(y)
    return np.array([mx, my, mz]).T


def homogeneous(xyz, theta, phi):
    """
    Initializes homogeneous magnetization state.

    Parameters
    __________

    xyz : coordinates of the nodes (sample vector)
    theta, phi : the required direction in spherical coordinates

    Return
    ------
    mag : Normalized magnetization in sample vector format.

    """
    theta = theta / 180.0 * np.pi
    phi = phi / 180.0 * np.pi
    mx = np.sin(theta) * np.cos(phi)
    my = np.sin(theta) * np.sin(phi)
    mz = np.cos(theta)
    return np.array([mx, my, mz], like=xyz)


def vortex(xyz, chi, p):
    """
     vortex state
    """
    x = xyz.T[0]
    y = xyz.T[1]

    rho_c = 10
    rho = np.sqrt(x ** 2 + y ** 2)

    # Usov Pechany ansatz
    theta = np.pi / 2 * np.ones_like(rho)
    theta[rho <= rho_c] = 2 * np.arctan(rho / rho_c)

    mx = chi * np.sin(theta) * (-np.sin(theta))
    my = chi * np.sin(theta) * (np.cos(theta))
    mz = p * np.cos(theta) * np.ones_like(x)
    return np.array([mx, my, mz]).T


def onion_tube_axial(xyz, pol_y=1, pol_n=1, pol_s=1, delta=5):
    """
    Onion state with domain wall along y at x = 0. The domain walls rotating on a cylinder (not in radial direction).

    pol_y : polarization of state in y direction (1 or -1)
    delta : thickness of domain walls (default is 5)
    pol_n : polarity (z) of domain wall in y > 0 region
    pol_s : polarity (z) of domain wall in y < 0 region
    """

    x = xyz.T[0]
    y = xyz.T[1]

    phi = np.arctan2(y, x)
    mx = np.zeros_like(x)
    my = np.zeros_like(x)
    # theta = np.ones_like(x)*np.pi/2

    theta = 2 * np.arctan(np.exp(x / delta)) - np.pi / 2

    mx[x > 0] = pol_y * (-np.sin(phi[x > 0])) * np.sin(theta[x > 0])
    mx[x <= 0] = pol_y * (-np.sin(phi[x <= 0])) * np.sin(theta[x <= 0])
    my[x > 0] = pol_y * (np.cos(phi[x > 0])) * np.sin(theta[x > 0])
    my[x <= 0] = pol_y * (np.cos(phi[x <= 0])) * np.sin(theta[x <= 0])

    mz = np.cos(theta)
    mz[y > 0] *= pol_n
    mz[y <= 0] *= pol_s
    return np.array([mx, my, mz]).T


def neel_tube(xyz, chi):
    x = xyz.T[0]
    y = xyz.T[1]
    phi = np.arctan2(y, x)
    mx = np.zeros_like(x)
    my = np.zeros_like(x)
    delta = 5
    # theta = np.ones_like(x)*np.pi/2
    theta = 2 * np.arctan(np.exp(x / delta))

    mx[x > 0] = chi * (-np.sin(phi[x > 0])) * np.sin(theta[x > 0])
    mx[x <= 0] = -chi * (-np.sin(phi[x <= 0])) * np.sin(theta[x <= 0])
    my[x > 0] = chi * (np.cos(phi[x > 0])) * np.sin(theta[x > 0])
    my[x <= 0] = -chi * (np.cos(phi[x <= 0])) * np.sin(theta[x <= 0])

    mz = np.cos(theta)
    return np.array([mx, my, mz]).T
